/*
	   Copyright (C) 2012 Gilles Landais (CDS)

    Author: Gilles Landais, Strasbourg astronomical Data Center (CDS)
    Email: gilles.landais@unistra.fr

    This file is part of H3C.

    H3C is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    H3C is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with H3C; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

/** @file h3c_util.c
 *  @brief usefull functions in the library heaplix for PostgreSQL
 *  @author G.Landais (CDS)
 *  @date 19/jul/2011
 *
 *  @note About the coordinate system,
 *  the Healpix library ofeered by the nasa deals with (theta,phi):
 *
 *  - theta: The polar angle or colatitude on the sphere, 
 *           ranging from 0 at the North Pole to PI at the South Pole.
 *  - phi  : The azimuthal angle on the sphere, in [0,2PI[.
 */
#include <math.h>
#include <sys/time.h>
#include <time.h>

#include "common.h"
#include "h3c_util.h"
#include "h3c_math.h"
#include "h3cpp_healpix.h"

/* PostgreSQL stuff */
#include <postgres.h>



/*****************************************************************************/
/** @brief print statistics
 *  @param level INFO
 *  @param fmt format
 *  @param ... optional argument used in format
 */
/*****************************************************************************/
void h3c_log(int level, char *fmt, ...)
{
#if _H3C_MUTE
    if (level == H3C_ERROR) {
        va_list l;
        char str[BUFSIZ];
  
        va_start(l, fmt);
        vsprintf(str, fmt, l);
#ifdef _H3C_MAIN
        fprintf(stderr, "(ERROR)H3C: %s\n", str);
        exit(1);
#else
        elog(ERROR, str);
#endif
    }
#else
    static struct timeval tv;
    static struct timezone tz;
    static int t0 = -1;
    int t;
    va_list l;
    char str[BUFSIZ];
  
    va_start(l, fmt);
    vsprintf(str, fmt, l);

    if (level < 0 || t0 == -1) {
#ifdef _H3C_MAIN
        fprintf(stderr, "(debug)H3C: begin time stat\n");
#else
        elog(DEBUG1, "H3C: begin time stat");
#endif
        gettimeofday(&tv, &tz);
        t0 = tv.tv_sec*1000000+tv.tv_usec;
        if (level < 0) return;
    }

    gettimeofday(&tv, &tz);
    t = (tv.tv_sec*1000000+tv.tv_usec)-t0;
#ifdef _H3C_MAIN
    fprintf(stderr, "(%s)H3C: %s [%.3f msec]\n",
        (level==H3C_DEBUG)?"debug":(level==H3C_INFO)?"info":"error",
        str, t/1000.);
    if (level == H3C_ERROR) exit(1);
#else
    elog(level, "H3C: %s [%.3f msec]",
        str, t/1000.);
#endif
    va_end(l);
#endif
}

/*****************************************************************************/
/** @brief initalize the debugger timer
 */
/*****************************************************************************/
void h3c_debug_init_clock() 
{
    return h3c_log(-1, NULL);
}

/*****************************************************************************/
/** @brief get the declination from the theta value
 *  @param theta
 *  @return declination
 */
/*****************************************************************************/
h3c_coord_t h3c_theta2dec(h3c_coord_t theta)
{
    return H3C_PI_2-theta;
}

/*****************************************************************************/
/** @brief get the theta value from the declination
 *  @param dec declination
 *  @return theta
 */
/*****************************************************************************/
h3c_coord_t h3c_dec2theta(h3c_coord_t dec)
{
    if (dec>H3C_PI_2) dec=H3C_PI_2;
    else if(dec<-1*H3C_PI_2) dec=-1*H3C_PI_2;
    return H3C_PI_2-dec;
}

/*****************************************************************************/
/** @brief get the right ascension from the phi value
 *  @param phi
 *  @return right ascension
 */
/*****************************************************************************/
h3c_coord_t h3c_phi2ra(h3c_coord_t phi)
{
    if (phi > H3C_PI) return phi-H3C_2PI;
    return phi;
}

/*****************************************************************************/
/** @brief get the phi value from the right ascension
 *  @param ra right ascension
 *  @return phi
 */
/*****************************************************************************/
h3c_coord_t h3c_ra2phi(h3c_coord_t ra)
{
    if (ra < 0) return H3C_2PI+ra;
    return ra;
}

/*****************************************************************************/
/** @brief get the order from the nside number
 *  @param nside nside level
 *  @return the order
 */
/*****************************************************************************/
int h3c_order(int nside) {
    unsigned int res=0;
    if (nside <= 0 || (nside & (nside-1))) {
        h3c_log(H3C_ERROR, "unvalid nside");
        return -1; 
    }

    while (nside > 0x0000FFFF) { res+=16; nside>>=16; }
    if (nside > 0x000000FF) { res|=8; nside>>=8; }
    if (nside > 0x0000000F) { res|=4; nside>>=4; }
    if (nside > 0x00000003) { res|=2; nside>>=2; }
    if (nside > 0x00000001) { res|=1; }
    return res;
}

/*****************************************************************************/
/** @brief get the number of ipix in the entire sphere
 *  @param nside nside level
 *  @return number of ipix in the entire sphere
 */
/*****************************************************************************/
h3c_ipix_t h3c_npix(int nside) {
    return 12*nside*nside;
}

/*****************************************************************************/
/** @brief get the max radius of ipix for a given nside
 *  @param nside nside level
 *  @return the radius
 */
/*****************************************************************************/
h3c_coord_t h3c_radius(int nside) {
    _h3c_vector a, b, c;
    double t;
    h3c_unit_vector(2./3., H3C_PI/(4*nside), a);
    t = 1.-(1./nside);
    t *= t;
    h3c_unit_vector(1-t/3., 0, b);
    h3c_vect_prod(a, b, c);

    return atan2 (h3c_vect_length(c), h3c_scalar_product(a, b))*H3C_RADEG;
}

/*****************************************************************************/
/** @brief (c++ interface) get the ipix list of a cone
 *  @param ra right ascension
 *  @param dec declination
 *  @param radius radius of the cone
 *  @param nside nside level
 *  @return ipix list
 */
/*****************************************************************************/
h3c_ipix_t *h3c_disk_ipix(h3c_coord_t ra,
                          h3c_coord_t dec,
                          h3c_coord_t radius,
                          int nside,
                          int *count) 
{
    return h3cpp_disk_ipix(ra, dec, radius, nside, count, 1);
}

#if 0
/* not effiscient because get too much ipix */

#define H3C_MAX_MEM_IPIX 6

/*****************************************************************************/
void h3c_sorted_list_insert(h3c_ipix_t *list, int count, h3c_ipix_t ipix) {
    int i;
    list[count] = ipix;
    for (i = count-1; i > -1; i--) {
        if (list[i] < ipix) return;
        list[i+1] = list[i];
        list[i] = ipix;
    }
}

/*****************************************************************************/
/* idea: get the ipix corresponing to the coordinates and search recursively
 *       all ipix arround (according to the property: 
 *       dist(MO) < radius+h3c_radius(nside) (O=(ra,dec), M=centre of the ipix)
 */
h3c_ipix_t *h3c_disk_ipix_1(h3c_coord_t ra,
                          h3c_coord_t dec,
                          h3c_coord_t radius,
                          int nside,
                          int *count)
{
    h3c_ipix_t ipix0;
    static h3c_ipix_t ipix[H3C_MAX_MEM_IPIX]; /* should never over H3C_MAX_MEM_IPIX */
    h3c_ipix_t *near; /* neighbors */
    h3c_coord_t theta,phi, ra0, dec0;
    int i;

    /* get the ipix where is the center of the disk */
    h3c_ang2pix_nest(nside, 
                     h3c_dec2theta(dec*H3C_DEGRA),
                     h3c_ra2phi(ra*H3C_DEGRA),
                     (h3c_ipix_t *) &ipix0);
    ipix[0] = ipix0;
    *count = 1;
    
    near = h3cpp_get_neighbors(ipix0, nside);
    for (i = 0; i < 8 ; i++) {
        h3c_pix2ang_nest(nside,
                         near[i],
                         &theta,
                         &phi);
        dec0 = h3c_theta2dec(theta)*H3C_RADEG;
        ra0 = h3c_phi2ra(phi)*H3C_RADEG;
        if (h3c_dist(ra, dec, ra0, dec0) < radius) {
            h3c_log(H3C_DEBUG,"ADD ipix %lld", near[i]);
            if ((*count) > H3C_MAX_MEM_IPIX-1) h3c_log(H3C_ERROR, "too much ipix!");
            h3c_sorted_list_insert(ipix, *count, near[i]);
            (*count)++;
        }
    }

    *count = (*count > 4) ? 4: *count; /* some ipix are forgotten !!!*/
    return ipix;
}
#endif

/*****************************************************************************/
/*****************************************************************************/
static int h3c_ipix_t_cmp(const void *a, const void *b) {
    h3c_ipix_t t =(*(h3c_ipix_t*)a)-*((h3c_ipix_t*)b);
    if (t==0) return 0;
    if (t>0) return 1;
    return -1;
/*    return (*(h3c_ipix_t*)a)-*((h3c_ipix_t*)b);*/
}

/*****************************************************************************/
/** @brief split a sorted list of ipix into a sorted list of segments.
 *  ex: {1,2,3,4,9,12,13,14} -> {1,4,9,9,12,14}
 */
/*****************************************************************************/
static int h3c_split_list(h3c_ipix_t *in,
                          int nin,
                          h3c_ipix_t *out,
                          int nout,
                          int step)
{
    int i, j = 0;
    out[0] = in[0];
    for (i = 1; i < nin; i++) {
        if (in[i]-in[i-1] > step) {
            j += 2;
            if (j < nout) {
                out[j-1] = in[i-1];
                out[j] = in[i];
            }
            else return -1;
        }
    }
    if (j+1 < nout) out[++j] = in[i-1];
    else return -1;
    return j+1;
}

/*****************************************************************************/
/** @brief get the ipix list of a cone and group them 
 *  @param ra right ascension
 *  @param dec declination
 *  @param radius radius of the cone
 *  @param nside nside level
 *  @param list the list of ipix to return (OUT)
 *  @param listsize size of list
 *  @return number of element of the ipix list
 *
 *  @note the list returned is sorted and grouped
 *
 *        ex: list=[1,3,4,4,6,8] => {1-3,4,6-8}
 *        list[2n]=min of a subset, list[2n+1]=max of a subset
 */
/*****************************************************************************/
#ifndef LIMIT_CIRCLE_ORDER_0
#define LIMIT_CIRCLE_ORDER_0 2/60.
#endif
#ifndef LIMIT_CIRCLE_ORDER_1
#define LIMIT_CIRCLE_ORDER_1 0.5
#endif
#ifndef LIMIT_CIRCLE_ORDER_2
#define LIMIT_CIRCLE_ORDER_2 1
#endif
#ifndef LIMIT_CIRCLE_ORDER_3
#define LIMIT_CIRCLE_ORDER_3 15
#endif
#ifndef LIMIT_CIRCLE_ORDER_4
#define LIMIT_CIRCLE_ORDER_4 40
#endif

int h3c_disk_ipix_it(h3c_coord_t ra,
                    h3c_coord_t dec,
                    h3c_coord_t radius,
                    int nside,
                    h3c_ipix_t *list,
                    int listsize)
{
    int j, count, step, gap, nside0 = nside;
    h3c_ipix_t *all;
    h3c_log(H3C_DEBUG, "begin getting ipix %d", nside);

#if 0
    all = h3c_query_disc(ra, dec, radius, nside, &count);
#else
    /* redefine nside in function of radius */
    if (radius >= LIMIT_CIRCLE_ORDER_0) {
        if (radius >= LIMIT_CIRCLE_ORDER_4) nside0 = nside>>6;
        else if (radius >= LIMIT_CIRCLE_ORDER_3) nside0 = nside>>5;
        else if (radius >= LIMIT_CIRCLE_ORDER_2) nside0 = nside>>4;
        else if (radius >= LIMIT_CIRCLE_ORDER_1)  nside0 = nside>>3;
        else nside0 = nside>>2;

        if (nside0 < 512) nside0 = 512;
        h3c_log(H3C_INFO,  "decrease nside to %d", nside0);
    }
    
    all = h3cpp_disk_ipix(ra, dec, radius, nside0, &count, 1);
#endif

    h3c_log(H3C_DEBUG, "end getting ipix");
    if (!all) {
        if (count > 0 ) h3c_log(H3C_ERROR, "index H3C uses too much ipix !");
        list[0] = -1;
        return 0;
    }

    qsort(all, count, sizeof(h3c_ipix_t), h3c_ipix_t_cmp);
    h3c_log(H3C_DEBUG, "end sorting");

    if (radius >= 1) { 
        if (radius >= 10) step = 256;
        else if (radius >= 5) step = 64;
        else step = 16;
    }
    else step = 4;
    gap = step;

    while ( (j = h3c_split_list(all, count, list, listsize, gap)) < 0 ) {
        gap += step;
        /*gap *= 2;*/
    }
    h3c_log(H3C_DEBUG, "end split");

    if (nside0 != nside)  {
        for (j=0; j<listsize; j++) {
            if (j%2) list[j] = (list[j]+1)<<((h3c_order(nside)-h3c_order(nside0))*2);
            else list[j] = (list[j])<<((h3c_order(nside)-h3c_order(nside0))*2);
        }
    }

    free(all);

#if _H3C_MUTE
#else
{   int i, e = 0;
    static int max = 0;
    static long long sum = 0;
    static int npass;

    for (i = 0; i < j/2; i++) e += list[2*i+1]-list[2*i]+1;
    h3c_log(H3C_DEBUG, "STAT (h3c_disk_ipix_it) #ipix=%d =split=> %d (gap=%d)",
        count, e, gap);

    npass++;
    sum += count;
    max = (count>max) ? count : max;
    if (npass % 10000 == 0) h3c_log(H3C_INFO, "npass=%d moy(#ipix)=%f max(#ipix)=%d", npass, (1.*sum/npass), max);
}
#endif
    return j;
}


/*****************************************************************************/
/** @brief calculate the distance between 2 points
 *  @param ra1 right ascention of the first point
 *  @param dec1 declination of the first point
 *  @param ra2 right ascention of the second point
 *  @param dec2 declination of the second point
 *  @return the distance
 *
 *  @note Use the haversine function
 */
/*****************************************************************************/
h3c_coord_t h3c_dist(h3c_coord_t ra1,
                     h3c_coord_t dec1,
                     h3c_coord_t ra2,
                     h3c_coord_t dec2)
{
    h3c_coord_t ra1Rad  = ra1*H3C_DEGRA;
    h3c_coord_t ra2Rad  = ra2*H3C_DEGRA;
    h3c_coord_t dec1Rad = dec1*H3C_DEGRA;
    h3c_coord_t dec2Rad = dec2*H3C_DEGRA;
    
    h3c_coord_t d = pow(sin((dec1Rad-dec2Rad)/2.),2);
    d += pow(sin((ra1Rad-ra2Rad)/2.), 2)*cos(dec1Rad)*cos(dec2Rad);
    return (2.*asin(sqrt(d)))*H3C_RADEG;
}

/*****************************************************************************/
/** @brief calculate the sinus distance (to be compatible with Q3C)
 *  @param ra1 right ascention of the first point
 *  @param dec1 declination of the first point
 *  @param ra2 right ascention of the second point
 *  @param dec2 declination of the second point
 *  @return the distance
 *
 *  @note Use the haversine function
 */
/*****************************************************************************/
h3c_coord_t h3c_sindist(h3c_coord_t ra1,
                        h3c_coord_t dec1,
                        h3c_coord_t ra2,
                        h3c_coord_t dec2)
{
    h3c_coord_t ra1Rad  = ra1*H3C_DEGRA;
    h3c_coord_t ra2Rad  = ra2*H3C_DEGRA;
    h3c_coord_t dec1Rad = dec1*H3C_DEGRA;
    h3c_coord_t dec2Rad = dec2*H3C_DEGRA;
    
    h3c_coord_t d = pow(sinl((dec1Rad-dec2Rad)/2.),2);
    d += pow(sinl((ra1Rad-ra2Rad)/2.), 2)*cosl(dec1Rad)*cos(dec2Rad);
    return d ;
}

/*****************************************************************************/
/*****************************************************************************/
static h3c_coord_t h3c_polygon_radius(h3c_coord_t *in_ra,
                                      h3c_coord_t *in_dec,
                                      int n)
{
    int i;
    h3c_coord_t center[2];
    h3c_coord_t d, dist = 0;

    h3c_poly_center(n, in_ra, in_dec, center);
    for (i=0; i<n; i++) {
        d = h3c_dist(center[0], center[1], in_ra[i], in_dec[i]);
        if (d > dist) dist = d;
    }
    return dist;
}

/*****************************************************************************/
/** @brief get the ipix list of a polygone and group them
 *  @param in_ra
 *  @param in_dec
 *  @param n number of points
 *  @param nside nside level
 *  @param list the ipix list to fill
 *  @param number of element of the ipix list
 *  @return number of elements filled
 *
 *  @note enclose the polygon inside a box
 */
/*****************************************************************************/
#ifndef LIMIT_POLYGONE_ORDER_0
#define LIMIT_POLYGONE_ORDER_0 1/60.
#endif
#ifndef LIMIT_POLYGONE_ORDER_1
#define LIMIT_POLYGONE_ORDER_1 0.5
#endif
#ifndef LIMIT_POLYGONE_ORDER_2
#define LIMIT_POLYGONE_ORDER_2 1
#endif

int h3c_polygon_ipix_it(h3c_coord_t *in_ra,
                            h3c_coord_t *in_dec,
                            int n,
                            int nside,
                            h3c_ipix_t *list,
                            int listsize) 
{
    int j, count, gap, step;
    h3c_ipix_t *all;
    int nside0 = nside;

    h3c_coord_t sz = h3c_polygon_radius(in_ra, in_dec, n);
h3c_log(H3C_INFO, "size=%f",sz);
    if (sz >= LIMIT_POLYGONE_ORDER_0) {
        if (sz >= LIMIT_POLYGONE_ORDER_2) nside0 = nside>>4;
        else if (sz >= LIMIT_POLYGONE_ORDER_1) nside0 = nside>>3;
        else nside0 = nside>>2;

        if (nside0 < 64) nside0 = 64;
        h3c_log(H3C_INFO, "decrease nside to %d (circumscribed circle r=%g)",
                nside0, sz);
    }

    h3c_log(H3C_DEBUG, "begin getting ipix");
    all = h3c_polygon_ipix(in_ra, in_dec, n, nside0, &count);
    if (!all) {
        list[0] = 0;
        h3c_log(H3C_INFO,"no healpix found ?..");
        list[1] = h3c_npix(nside);
        return 2;
    }

    qsort(all, count, sizeof(h3c_ipix_t), h3c_ipix_t_cmp);
    h3c_log(H3C_DEBUG, "end sorting (%d items)", count);

    if (count > 1000000) step = 4096;
    else if (count > 100000) step = 1024;
    else if (count > 10000) step = 32;
    else step = 4;
    gap = step;
     
    while ( (j = h3c_split_list(all, count, list, listsize, gap))<0 ) {
        gap += step;
    }
    h3c_log(H3C_DEBUG, "end split");

    free(all);

    if (nside0 != nside)  {
        int i;
        for (i=0; i<listsize/2; i++) {
             list[2*i+1] = (list[2*i+1]+1)<<((h3c_order(nside)-h3c_order(nside0))*2);
             list[2*i] = (list[2*i])<<((h3c_order(nside)-h3c_order(nside0))*2);
        }
    }

#if _H3C_MUTE
#else
{   int i, e = 0;
    for (i = 0; i < j/2; i++) e += (list[2*i+1]-list[2*i])+1;
    h3c_log(H3C_DEBUG, "STAT (h3c_polygon_ipix_it) #ipix=%d =split=> %d (gap=%d)",
        count, e, gap);
}
#endif
    return j;
}

/*****************************************************************************/
/** @brief verify if a point is inside an ellipse
 *
 *  @note: code coming from Q3C
 */
/*****************************************************************************/
char h3c_in_ellipse(h3c_coord_t alpha, h3c_coord_t delta0,
        h3c_coord_t alpha1, h3c_coord_t delta01, h3c_coord_t d0,
        h3c_coord_t e, h3c_coord_t PA0)
{
        h3c_coord_t d_alpha = (alpha1 - alpha) * H3C_DEGRA;
        h3c_coord_t delta1 = delta01 * H3C_DEGRA;
        h3c_coord_t delta = delta0 * H3C_DEGRA;
        h3c_coord_t PA = PA0 * H3C_DEGRA;
        h3c_coord_t d = d0 * H3C_DEGRA;
        
        h3c_coord_t t1 = cos(d_alpha);
        h3c_coord_t t22 = sin(d_alpha);
        h3c_coord_t t3 = cos(delta1);
        h3c_coord_t t32 = sin(delta1);
        h3c_coord_t t6 = cos(delta);
        h3c_coord_t t26 = sin(delta);
        h3c_coord_t t9 = cos(d);
        h3c_coord_t t55 = sin(d);
        
        h3c_coord_t t2;
        h3c_coord_t t4;
        h3c_coord_t t5;
        h3c_coord_t t7;
        h3c_coord_t t8;
        h3c_coord_t t10;
        h3c_coord_t t11;
        h3c_coord_t t13;
        h3c_coord_t t14;
        h3c_coord_t t15;
        h3c_coord_t t18;
        h3c_coord_t t19;
        h3c_coord_t t24;
        h3c_coord_t t31;
        h3c_coord_t t36;
        h3c_coord_t t37;
        h3c_coord_t t38;
        h3c_coord_t t45;
        
        h3c_coord_t t56;
        h3c_coord_t t57;
        h3c_coord_t t60;
        h3c_coord_t t61;
        h3c_coord_t t63;
        
        if ((t3 * t6 * t1 + t32 * t26) < 0)
        {
                return 0;
        }
        
        t2 = t1*t1;
        
        t4 = t3*t3;
        t5 = t2*t4;
        
        t7 = t6*t6;
        t8 = t5*t7;
        
        t10 = t9*t9;
        t11 = t7*t10;
        t13 = cos(PA);
        t14 = t13*t13;
        t15 = t14*t10;
        t18 = t7*t14;
        t19 = t18*t10;
        
        t24 = sin(PA);
        
        t31 = t1*t3;
        
        t36 = 2.0*t31*t32*t26*t6;
        t37 = t31*t32;
        t38 = t26*t6;
        t45 = t4*t10;
        
        t56 = t55*t55;
        t57 = t4*t7;
        t60 = -t8+t5*t11+2.0*t5*t15-t5*t19-2.0*t1*t4*t22*t10*t24*t13*t26-t36+2.0*t37*t38*t10-2.0*t37*t38*t15-t45*t14-t45*t2+2.0*t22*t3*t32*t6*t24*t10*t13-t56+t7-t11+t4-t57+t57*t10+t19-t18*t45;
        t61 = e*e;
        t63 = t60*t61+t8+t57-t4-t7+t56+t36;
        return t63 > 0;
}

/*****************************************************************************/
void h3c_pix2ang_nest(long nside, 
                      h3c_ipix_t ipnest,
                      h3c_coord_t *theta,
                      h3c_coord_t *phi) {
    h3cpp_pix2ang_nest(nside, ipnest, theta, phi);
}

/*****************************************************************************/
void h3c_ang2pix_nest(long nside,
                      h3c_coord_t theta,
                      h3c_coord_t phi,
                      h3c_ipix_t *ipnest) {
    h3cpp_ang2pix_nest(nside, theta, phi, ipnest);
}

/*****************************************************************************/
void h3c_ring2nest(long nside,
                   h3c_ipix_t ipring,
                   h3c_ipix_t *ipnest) {
    h3cpp_ring2nest(nside, ipring, ipnest);
}


